/*
Module : MEMMAP.CPP
Purpose: Implementation for an MFC class to wrap memory mapped files
Created: PJN / MEMMAP/1 / 19-03-1997
History: NONE


Copyright (c) 1997 by PJ Naughter.  
All rights reserved.

*/

/////////////////////////////////  Includes  //////////////////////////////////
#include "stdafx.h"
#include "memmap.h"



///////////////////////////////// Implementation //////////////////////////////



IMPLEMENT_DYNAMIC(CMemMapFile, CObject)


CMemMapFile::CMemMapFile()
{
  //Initialise variables to sane values
  m_hFile = INVALID_HANDLE_VALUE;
  m_hMapping = NULL;
  m_bReadOnly = TRUE;
  m_lpData = NULL;
  m_bOpen = FALSE;
  m_hMutex = NULL;
  m_dwLength = 0;
}


CMemMapFile::~CMemMapFile()
{
  UnMap();
}


BOOL CMemMapFile::MapFile(const CString& sFilename, BOOL bReadOnly, DWORD dwShareMode)
{
  //Work out the file access flags
  m_bReadOnly = bReadOnly;
	DWORD dwDesiredFileAccess = GENERIC_READ;
	if (!m_bReadOnly)
		dwDesiredFileAccess |= GENERIC_WRITE;

  //Open the real file on the file system
  m_hFile = CreateFile(sFilename, dwDesiredFileAccess, dwShareMode, NULL, 
                       OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (m_hFile == INVALID_HANDLE_VALUE)
  {
    UnMap();
    return FALSE;
  }

	// Get the size of the file we are mapping
	BY_HANDLE_FILE_INFORMATION bhInfo;
  BOOL bSuccess = GetFileInformationByHandle(m_hFile, &bhInfo);
	if (bSuccess && (bhInfo.nFileSizeHigh == 0))  //Currently only support up to 4GB size files
  {
  	m_dwLength = bhInfo.nFileSizeLow;
  }
  else
  {
    UnMap();
    return FALSE;
  }

  //Do the actual mapping
  m_sMappingName = CreateMappingName(sFilename);
  return MapHandle(m_hFile);
}


BOOL CMemMapFile::MapMemory(const CString& sName, DWORD dwBytes, BOOL bReadOnly)
{
  m_dwLength = dwBytes;
  m_bReadOnly = bReadOnly;
  m_sMappingName = CreateMappingName(sName);
  return MapHandle((HANDLE)0xFFFFFFFF);
}


BOOL CMemMapFile::MapExistingMemory(const CString& sName, DWORD dwBytes, BOOL bReadOnly)
{
  m_dwLength = dwBytes;
  m_bReadOnly = bReadOnly;

  DWORD dwDesiredAccess = (!bReadOnly) ? FILE_MAP_WRITE : FILE_MAP_READ;
  m_sMappingName = CreateMappingName(sName);
  m_hMapping = OpenFileMapping(dwDesiredAccess,	0, m_sMappingName);
  if (m_hMapping == NULL)
  {
    UnMap();
    return FALSE;
  }

  //Map the view
  m_lpData = MapViewOfFile(m_hMapping, dwDesiredAccess, 0, 0, m_dwLength);

  //Create the mutex to sync access
  m_hMutex = CreateMutex(NULL, FALSE, CreateMutexName());

  return (m_lpData != NULL);
}


BOOL CMemMapFile::MapHandle(HANDLE hHandle)
{
  //Create the file mapping object
  DWORD flProtect = (!m_bReadOnly) ? PAGE_READWRITE : PAGE_READONLY;
  m_hMapping = ::CreateFileMapping(hHandle, NULL, flProtect, 0, m_dwLength, m_sMappingName);
  if (m_hMapping == NULL)
  {
    UnMap();
    return FALSE;
  }

  //Map the view
	DWORD dwDesiredAccess = (!m_bReadOnly) ? FILE_MAP_WRITE : FILE_MAP_READ;
  m_lpData = MapViewOfFile(m_hMapping, dwDesiredAccess, 0, 0, 0);

  //Create the mutex to sync access
  m_hMutex = CreateMutex(NULL, FALSE, CreateMutexName());

  return (m_lpData != NULL);
}


LPVOID CMemMapFile::Open(DWORD dwTimeout)
{
  if (m_lpData == NULL)
    return NULL;

  //Synchronise access to the MMF using the named mutex
	DWORD dwResult = WaitForSingleObject(m_hMutex, dwTimeout);

	if (dwResult == WAIT_OBJECT_0)
	{
		m_bOpen = TRUE;
		return m_lpData;
	}

  return NULL;
}


BOOL CMemMapFile::Close()
{
  //Release our interest in this MMF
  if (!m_bOpen)
    return FALSE;

  m_bOpen = FALSE;
  ReleaseMutex(m_hMutex);
  
  return TRUE;
}


BOOL CMemMapFile::Flush()
{
  if (m_lpData == NULL)
		return FALSE;

  return FlushViewOfFile(m_lpData, 0);
}


CString CMemMapFile::GetMappingName() const
{
  return m_sMappingName;
}


CString CMemMapFile::CreateMutexName() const
{
  return m_sMappingName + _T("MUTEX");
}


void CMemMapFile::UnMap()
{
  //Close any views which may be open
	Close();

  //unmap the view
	if (m_lpData != NULL)
	{
		FlushViewOfFile(m_lpData, 0);
		UnmapViewOfFile(m_lpData);
		m_lpData = NULL;
	}

  //remove the file mapping
	if (m_hMapping != NULL)
	{
		CloseHandle(m_hMapping);
		m_hMapping = NULL;
	}

  //close the file system file if its open
	if (m_hFile != INVALID_HANDLE_VALUE)
	{
		CloseHandle(m_hFile);
		m_hFile = INVALID_HANDLE_VALUE;
	}

  //Close the mutex we have been using
	if (m_hMutex != NULL)
	{
		CloseHandle(m_hMutex);
		m_hMutex = NULL;
	}

  //Reset the remaining member variables
  m_bReadOnly = TRUE;
  m_sMappingName = _T("");
  m_dwLength = 0;
}


CString CMemMapFile::CreateMappingName(const CString& sName)
{
  CString rVal(sName);

  //Replace all '\' by '_'
  for (int i=0; i<sName.GetLength(); i++)
  {
    if (rVal.GetAt(i) == _T('\\'))
      rVal.SetAt(i, _T('_'));
  }

  rVal.MakeUpper();

  return rVal;
}


#ifdef _DEBUG
void CMemMapFile::AssertValid() const
{
  CObject::AssertValid();
}
#endif


#ifdef _DEBUG
void CMemMapFile::Dump(CDumpContext& dc) const
{
  CObject::Dump(dc);

  CString sText;
  sText.Format(_T("m_hFile=0x%08x\n"), m_hFile);
  dc << sText;
  sText.Format(_T("m_hMapping=0x%08x\n"), m_hMapping);
  dc << sText;
  sText.Format(_T("m_bReadOnly=%d\n"), m_bReadOnly);
  dc << sText;
  sText.Format(_T("m_lpData=0x%08x\n"), m_lpData);
  dc << sText;
  sText.Format(_T("m_sMappingName=%s\n"), m_sMappingName);
  dc << sText;
  sText.Format(_T("m_bOpen=%d\n"), m_bOpen);
  dc << sText;
  sText.Format(_T("m_hMutex=0x%08x\n"), m_hMutex);
  dc << sText;
  sText.Format(_T("m_dwLength=%d\n\n"), m_dwLength);
  dc << sText;
}
#endif



